#include "stdafx.h"
#include "_project.h"

#include "CDBExecutor.h"

namespace fusion = boost::fusion;

#ifndef USE_EXPORT_KEYWORD
#define doExport /*nothing*/
#endif

CDBExecutor::CDBExecutor(ConnectionManager* connManager, WinLogger *logger) : 
conn_(NULL),
connManager_(connManager),
logger_(logger),
errorHandler_(NULL)
{
	pg_mng_.reset( new Detail::DBManager(*connManager_) );
};
WinLogger* CDBExecutor::getLogger() const
{
	return logger_;
};
std::string CDBExecutor::getTypeName(bool cut) const
{
	if(!cut) return comp_type_;
	std::string t;
	t = (comp_type_.find(".") != std::string::npos) ? comp_type_.substr(comp_type_.find_first_of(".") + 1, comp_type_.length()) 
		: comp_type_;
	return t;
};
void CDBExecutor::cleanUp(void)
{
	tablename_ = comp_type_ = "";
	is_composite_ = false;
};
std::string CDBExecutor::getStatement(void) const
{
	return stmt_;
};
void CDBExecutor::setStatement(const std::string& s)
{
	stmt_ = s;
};
void CDBExecutor::adjustSQLCmd(const std::string &in, std::string &out)
{
	out = in;
	StringUtils::trim_left(out);
	StringUtils::trim_right(out);
};
std::string CDBExecutor::prepareAutoNameSQL(const std::string &sqlCmd)
{
	std::string stmtName;

	ERROR_HANDLER_START();
	connManager_->getUniquePrepStmtName(stmtName);
	prepareSQL(sqlCmd, stmtName);
	ERROR_HANDLER_END();

	return stmtName;
};
void CDBExecutor::prepareSQL(const std::string &sqlCmd, const std::string &prepKey)
{
	ERROR_HANDLER_START();
	prepareConnection();

	std::string adjustedSQLCmd;
	adjustSQLCmd(sqlCmd, adjustedSQLCmd);

	if (adjustedSQLCmd.length() == 0)
	{
		throw Smartdb::argument_error("SQL command is empty", SOURCELINE);
	}
	bool preparedCmd = isInPrepCmdRep(prepKey);

	if (preparedCmd)
	{
		deallocPreparedSQL(prepKey);
	}
	addToPrepCmdRep(adjustedSQLCmd, prepKey);
	const unsigned nparam = getNumPar(adjustedSQLCmd);

	pg_mng_->prepareQuery(prepKey, adjustedSQLCmd, nparam, true /*do escape*/);

	logger_->Log(logTRACE) << "Prepare: \"" << adjustedSQLCmd << "\" using: \"" << prepKey << "\" command name";

	ERROR_HANDLER_END();
};
const unsigned CDBExecutor::getNumPar( const std::string &sqlClause ) const
{
	const std::string &str = sqlClause;
	std::string::const_iterator i = str.begin();

	int num = 0;
	int max = 0;

	enum eStates {eNull, eDollar, eExtractNumber};

	eStates state = eNull;
	for(; i != str.end(); ++i)
	{
		if (*i == '$')
		{
			num = 0;
			state = eDollar;
			continue;
		}
		if (state == eNull)
			continue;
		if ( '0' <= *i && *i <= '9')
		{
			state = eExtractNumber;
			num = num * 10 + (*i - '0');
			continue;
		}
		if (state == eDollar)
		{
			state = eNull;
			continue;
		}
		if (state == eExtractNumber)
		{
			max = max > num ? max : num;
			state = eNull;
			continue;
		}
	}
	if (state == eExtractNumber)
	{
		max = max > num ? max : num;
		state = eNull;
	}
	return max;
};
CDBExecutor::~CDBExecutor()
{
	cleanUp();
	//clean up of notice message dispatcher
	NoticeDispatcher::unSubscribeForNotice(getSelfAsString());
};
void CDBExecutor::addParam(const char *paramType, const std::string* value, const boost::false_type&)//buildin types, will be send by value
{
	params_->putString(paramType, value);
}

void CDBExecutor::addParam(const char *paramType)
{
	if (!params_.get())
	{
		params_.reset(pg_mng_->getQueryParam());
	}
	params_->put_null();
};
/*
void CDBExecutor::addParam(const char *paramType, const PGvarchar& value)
{
	ERROR_HANDLER_START();
	prepareConnection();
	if (!getParam())
		createParam(getParam());

	int r = PQputf( getParam(),paramType, value);
	CHECKEQ(r, 0, SOURCELINE);
	ERROR_HANDLER_END();
};
void CDBExecutor::addParam(const char* paramType, const PGdate &value)
{
	ERROR_HANDLER_START();
	prepareConnection();
	if (!getParam())
		createParam(getParam());

	int r = PQputf( getParam(),paramType, &value);
	CHECKEQ(r, 0, SOURCELINE);
	ERROR_HANDLER_END();
};
void CDBExecutor::addParam(const char* paramType, const PGtimestamp &value)
{
	ERROR_HANDLER_START();
	prepareConnection();
	if (!getParam())
		createParam(getParam());

	int r = PQputf( getParam(),paramType, &value);
	CHECKEQ(r, 0, SOURCELINE);
	ERROR_HANDLER_END();
};
void CDBExecutor::addParam(const char* paramType, const PGmacaddr &value)
{
	ERROR_HANDLER_START();
	prepareConnection();
	if (!getParam())
		createParam(getParam());

	int r = PQputf( getParam(),paramType, &value);
	CHECKEQ(r, 0, SOURCELINE);
	ERROR_HANDLER_END();
};
void CDBExecutor::addParam(const char* paramType, const PGinet &value)
{
	ERROR_HANDLER_START();
	prepareConnection();
	if (!getParam())
		createParam(getParam());

	int r = PQputf( getParam(),paramType, &value);
	CHECKEQ(r, 0, SOURCELINE);
	ERROR_HANDLER_END();
};*/
void CDBExecutor::prepareConnection()
{
	if (!connManager_)
	{
		throw Smartdb::internal_error("Invalid reference  to \"connection manager\"", SOURCELINE);
	}
	conn_ = connManager_->getConnection();
	NoticeDispatcher::registerForNotice(getSelfAsString());
	//int res = PQregisterTypeHandler(conn_, "int" ,rgb_put, rgb_get);
	/*if (!res)
	{
	std::cout << PQgeterror() << std::endl;
	}*/
};
void CDBExecutor::deallocPreparedSQL(const std::string &prepKey)
{
	ERROR_HANDLER_START();
	bool res = isInPrepCmdRep(prepKey);
	if (!res)
		throw Smartdb::usage_error(boost::str(boost::format("No prepared command found for \"%1%\" command name ") % prepKey.c_str()), SOURCELINE);

	removeFromPrepCmdRep(prepKey);
	std::string cmd = DEALLOCATE_PREPARE;
	cmd += " \"";
	cmd += prepKey.c_str();
	cmd += "\"";
	logger_->Log(logTRACE) << "Deallocating \"" << prepKey << "\" prepared command.";
	
	pg_mng_->execSimpleQuery(cmd, 1);//todo remove magic number

	ERROR_HANDLER_END();
};
bool CDBExecutor::isInPrepCmdRep(const std::string &prepKey) const
{
	TPrepCmdRep::const_iterator i = prepCmdRep_.find(prepKey);
	if (i != prepCmdRep_.end())
	{
		return true;
	}
	return false;
};
void CDBExecutor::addToPrepCmdRep(const std::string &cmd, const std::string &prepKey)
{
	prepCmdRep_.insert(std::make_pair(prepKey, cmd));
};
bool CDBExecutor::removeFromPrepCmdRep(/*IN*/const std::string &prepKey)
{
	TPrepCmdRep::iterator i = prepCmdRep_.find(prepKey);
	if (i != prepCmdRep_.end())
	{
		prepCmdRep_.erase(i);
	}	
	else
	{
		logger_->Log(logWARNING) << "\"" << prepKey << "\"" << " key is not found in prepared statements repository";
	}
	return true;
};
bool CDBExecutor::getFromPrepCmdRep(/*IN*/const std::string &prepKey, /*OUT*/std::string &prepCmd)
{
	TPrepCmdRep::iterator i = prepCmdRep_.find(prepKey);
	if (i != prepCmdRep_.end())
	{
		prepCmd = (*i).second;
		return true;
	}
	return false;
};
/*
void CDBExecutor::execSQL(const std::string &cmd, int format/*= 1)
{
	ERROR_HANDLER_START();

	PQExecSQL(cmd, getParam(), SOURCELINE, format);
	clearParams();
	logger_->Log(logTRACE) << "Executing: " << cmd.c_str();
	ERROR_HANDLER_END();
};*/
void CDBExecutor::execSQL(const std::string &cmd, int format/*= 1*/)
{
	ERROR_HANDLER_START();
	
	if (params_.get())
	{
		pg_mng_->execParamQuery(*params_.get(), cmd, format);
		params_.reset(NULL);
	}
	else
	{
		pg_mng_->execSimpleQuery(cmd, format);
	}

	logger_->Log(logTRACE) << "Executing: " << cmd.c_str();

	ERROR_HANDLER_END();
};
void CDBExecutor::execPreparedSQL(const std::string &stmtName, int format)
{
	ERROR_HANDLER_START();
	prepareConnection();
	std::string prepCmd;
	bool res = getFromPrepCmdRep(stmtName, prepCmd);
	if (!res)
		throw Smartdb::usage_error(boost::str(boost::format("No prepared command found for \"%1%\" command name ") % stmtName.c_str()), SOURCELINE);

	setStatement(prepCmd);

	if (params_.get())
	{
		pg_mng_->execParamPrepared(*params_.get(), stmtName, format);
	}
	else
	{
		pg_mng_->execPrepared(stmtName, format);
	}

	ERROR_HANDLER_END();
};
void CDBExecutor::clearParams()
{
	//todo delete correctly parameters
	params_.reset(NULL);
};
void CDBExecutor::setCustomNoticeReceiver(const UsrNoticeReceiver proc) const
{
	ERROR_HANDLER_START();
	NoticeDispatcher::subcribeForNotice(getSelfAsString(), proc);
	ERROR_HANDLER_END();
};
void CDBExecutor::clearNoticeReceiver() const
{
	setCustomNoticeReceiver(NULL);
};
std::string CDBExecutor::getSelfAsString(void) const
{
	std::stringstream buf;
	buf << this;
	return buf.str();
};