#pragma once
#include "_project.h"
#include "PG_LibPQTypes.h"

namespace Smartdb
{

class PG_LIBPQTYPES_API std::logic_error;
class PG_LIBPQTYPES_API std::runtime_error;
class PG_LIBPQTYPES_API std::invalid_argument;
class PG_LIBPQTYPES_API std::domain_error;
class PG_LIBPQTYPES_API std::out_of_range;

/** 
* Abstract base class for all smart db exceptions
*/
class PG_LIBPQTYPES_API smartdb_exception
{
	std::string errPointOfError_;
public:
	explicit smartdb_exception(const std::string &errPoint);
	virtual ~smartdb_exception() throw () =0;
	/**
	* Returns point of error
	*/
	virtual const std::string getErrPoint() const { return errPointOfError_; };
	/* 
	* Method to retrive instance of exeption class
	* Example:
	* \code
	* try
    * {
    *   //smartdb calls
    * }
    * catch (const pqxx::pqxx_exception &e)
    * {
    *   std::cerr << e.getErrPoint() << << e.base().what() << std::endl;
	*   //for casting to a type of error
    *   const pqxx::sql_error *s=dynamic_cast<const pqxx::sql_error*>(&e.base());
    *   if (s) std::cerr << "Query was: " << s->query() << std::endl;
    * }
	* \endcode
	*/
	virtual const std::exception &base() const throw () = 0;

};
/**
* Run-time failure encountered by smartdb, similar to std::runtime_error
*/
class PG_LIBPQTYPES_API  failure :
  public smartdb_exception, public std::runtime_error
{
  virtual const std::exception &base() const throw () { return *this; }
public:
  failure(const std::string &err, const std::string &errPoint);
};

/**
* Exception class for protocol violation failure.
*/

class PG_LIBPQTYPES_API  protocol_violation : public failure
{
public:
  protocol_violation();
  protocol_violation(const std::string &err, const std::string &errPoint);
};

/**
* Exception class for lost or failed backend connection.
*/
class PG_LIBPQTYPES_API  broken_connection : public failure
{
public:
  broken_connection();
  broken_connection(const std::string &err, const std::string &errPoint);
};

/** 
* Exception class for failed queries.
*/
class PG_LIBPQTYPES_API  sql_error : public failure
{
public:
  sql_error();
  sql_error(const std::string &err, const std::string &errPoint);
  virtual ~sql_error() throw ();
};

/** Exception that might be thrown in rare cases where the connection to the
 * database is lost while finishing a database transaction, and there's no way
 * of telling whether it was actually executed by the backend.  In this case
 * the database is left in an indeterminate (but consistent) state, and only
 * manual inspection will tell which is the case.
 */
class PG_LIBPQTYPES_API  in_doubt_error : public failure
{
public:
  in_doubt_error(const std::string &err, const std::string &errPoint);
};

/**
* Internal error in smartdb library
*/
class PG_LIBPQTYPES_API  internal_error :
  public smartdb_exception, public std::logic_error
{
  virtual const std::exception &base() const throw () { return *this; }
public:
	internal_error(const std::string &err, const std::string &errPoint);
};

/**
* Error in usage of smartdb library, similar to std::logic_error
*/
class PG_LIBPQTYPES_API  usage_error :
  public smartdb_exception, public std::logic_error
{
  virtual const std::exception &base() const throw () { return *this; }
public:
	usage_error(const std::string &err, const std::string &errPoint) : logic_error(err), smartdb_exception(errPoint) {};
};

/**
* Invalid argument passed to smartdb, similar to std::invalid_argument
*/
class PG_LIBPQTYPES_API  argument_error :
  public smartdb_exception, public std::invalid_argument
{
  virtual const std::exception &base() const throw () { return *this; }
public:
	argument_error(const std::string &err, const std::string &errPoint) : invalid_argument(err), smartdb_exception(errPoint) {};
};

/**
* Convertion errors in smartdb, similar to std::domain_error
*/
class PG_LIBPQTYPES_API  conversion_error :
  public smartdb_exception, public std::domain_error
{
  virtual const std::exception &base() const throw () { return *this; }
public:
	conversion_error(const std::string &err, const std::string &errPoint) : domain_error(err), smartdb_exception(errPoint) {};
};

/**
* Out of range smartdb error, similar to std::out_of_range
*/
class PG_LIBPQTYPES_API  range_error :
  public smartdb_exception, public std::out_of_range
{
  virtual const std::exception &base() const throw () { return *this; }
public:
	range_error(const std::string &err, const std::string &errPoint) : std::out_of_range(err), smartdb_exception(errPoint) {};
};

/**
* Database feature not supported in current setup
*/
class PG_LIBPQTYPES_API  feature_not_supported : public sql_error
{
public:
	feature_not_supported(const std::string &err, const std::string &errPoint) : sql_error(err, errPoint) {};
};

/** 
* Error in data provided to SQL statement
*/
class PG_LIBPQTYPES_API  data_exception : public sql_error
{
public:
	data_exception(const std::string &err, const std::string &errPoint) : sql_error(err, errPoint) {};
	
};

/** 
* SQL integrity constraint violation smartdb error
*/
class PG_LIBPQTYPES_API  integrity_constraint_violation : public sql_error
{
public:
  integrity_constraint_violation(const std::string &err, const std::string &errPoint) :
	sql_error(err, errPoint) {}
  
};

/** 
* SQL restriction violation smartdb error
*/
class PG_LIBPQTYPES_API  restrict_violation :
  public integrity_constraint_violation
{
public:
  restrict_violation(const std::string &err, const std::string &errPoint) :
	integrity_constraint_violation(err, errPoint) {};
};

/** 
* SQL not unll violation smartdb error
*/
class PG_LIBPQTYPES_API  not_null_violation :
  public integrity_constraint_violation
{
public:
  not_null_violation(const std::string &err, const std::string &errPoint) :
	integrity_constraint_violation(err, errPoint) {};
};

/** 
* SQL foreign key violation smartdb error
*/
class PG_LIBPQTYPES_API  foreign_key_violation :
  public integrity_constraint_violation
{
public:
  foreign_key_violation(const std::string &err, const std::string &errPoint) :
	integrity_constraint_violation(err, errPoint) {}
};

/** 
* SQL unique violation smartdb error
*/
class PG_LIBPQTYPES_API  unique_violation :
  public integrity_constraint_violation
{
public:
  unique_violation(const std::string &err, const std::string &errPoint) :
	integrity_constraint_violation(err, errPoint) {}
  
};

/** 
* SQL check violation smartdb error
*/
class PG_LIBPQTYPES_API  check_violation :
  public integrity_constraint_violation
{
public:
  check_violation(const std::string &err, const std::string &errPoint) :
	integrity_constraint_violation(err, errPoint) {}
};

/** 
* SQL invalid cursor state smartdb error
*/
class PG_LIBPQTYPES_API  invalid_cursor_state : public sql_error
{
public:
  invalid_cursor_state(const std::string &err, const std::string &errPoint) : sql_error(err, errPoint) {}
};

/** 
* SQL invalid sql statement name smartdb error
*/
class PG_LIBPQTYPES_API  invalid_sql_statement_name : public sql_error
{
public:
  invalid_sql_statement_name(const std::string &err, const std::string &errPoint) :
	sql_error(err, errPoint) {}
};

/** 
* SQL invalid cursor name smartdb error
*/
class PG_LIBPQTYPES_API  invalid_cursor_name : public sql_error
{
public:
  invalid_cursor_name(const std::string &err, const std::string &errPoint) : sql_error(err, errPoint) {}
};

/** 
* SQL invalid schema name smartdb error
*/
class PG_LIBPQTYPES_API  invalid_schema_name : public sql_error
{
public:
  invalid_schema_name(const std::string &err, const std::string &errPoint) : sql_error(err, errPoint) {}
};

/** 
* SQL syntax smartdb error
*/
class PG_LIBPQTYPES_API  syntax_error : public sql_error
{
public:
  syntax_error(const std::string &err, const std::string &errPoint) : sql_error(err, errPoint) {}
};

/** 
* SQL undefined column smartdb error
*/
class PG_LIBPQTYPES_API  undefined_column : public syntax_error
{
public:
  undefined_column(const std::string &err, const std::string &errPoint) : syntax_error(err, errPoint) {}
};

/** 
* SQL undefined function smartdb error
*/
class PG_LIBPQTYPES_API  undefined_function : public syntax_error
{
public:
  undefined_function(const std::string &err, const std::string &errPoint) : syntax_error(err, errPoint) {}
};

/** 
* SQL undefined table smartdb error
*/
class PG_LIBPQTYPES_API  undefined_table : public syntax_error
{
public:
  undefined_table(const std::string &err, const std::string &errPoint) : syntax_error(err, errPoint) {}
};

/** 
* SQL insufficient privilege smartdb error
*/
class PG_LIBPQTYPES_API  insufficient_privilege : public sql_error
{
public:
  insufficient_privilege(const std::string &err, const std::string &errPoint) : sql_error(err, errPoint) {}
};

/** 
* SQL datatype mismatch smartdb error
*/
class PG_LIBPQTYPES_API datatype_mismatch : public sql_error
{
public:
  datatype_mismatch(const std::string &err, const std::string &errPoint) : sql_error(err, errPoint) {}
};

/** 
* SQL insufficient resources on server, smartdb error
*/
class PG_LIBPQTYPES_API  insufficient_resources : public sql_error
{
public:
  insufficient_resources(const std::string &err, const std::string &errPoint) : sql_error(err, errPoint) {}
};

/** 
* SQL disk full, smartdb error
*/
class PG_LIBPQTYPES_API  disk_full : public insufficient_resources
{
public:
  disk_full(const std::string &err, const std::string &errPoint) : insufficient_resources(err, errPoint) {}
};

/** 
* SQL out of memory, smartdb error
*/
class PG_LIBPQTYPES_API  out_of_memory : public insufficient_resources
{
public:
  out_of_memory(const std::string &err, const std::string &errPoint) :
	insufficient_resources(err, errPoint) {}
};

/** 
* SQL too many connections, smartdb error
*/
class PG_LIBPQTYPES_API  too_many_connections : public broken_connection
{
public:
  too_many_connections(const std::string &err, const std::string &errPoint) :
	broken_connection(err, errPoint) {}
};
/**
* Checks connection for sql errors
* @param conn connection to check
* @param res query result
* @param errPoint point of error. Thrown exception(if found) will be initializet with this error point.
*/
void PG_LIBPQTYPES_API CheckStatus(PGconn *conn, const PGresult *res, const std::string &errPoint);

}